# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ... import core
from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.http_response import AsyncHttpResponse, HttpResponse
from ...core.jsonable_encoder import jsonable_encoder
from ...core.request_options import RequestOptions
from ...core.unchecked_base_model import construct_type
from ..errors.bad_request_error import BadRequestError
from ..errors.forbidden_error import ForbiddenError
from ..errors.internal_server_error import InternalServerError
from ..errors.not_found_error import NotFoundError
from ..errors.unauthorized_error import UnauthorizedError
from .types.delete_audience_id_members_response import DeleteAudienceIdMembersResponse
from .types.delete_audience_id_response import DeleteAudienceIdResponse
from .types.get_audience_id_members_response import GetAudienceIdMembersResponse
from .types.get_audience_id_members_search_response import GetAudienceIdMembersSearchResponse
from .types.get_audience_id_response import GetAudienceIdResponse
from .types.get_audience_response import GetAudienceResponse
from .types.post_audience_id_members_response import PostAudienceIdMembersResponse
from .types.post_audience_response import PostAudienceResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawAudienceClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def get_all_audiences(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[GetAudienceResponse]:
        """
        Retrieve all audiences created by the authenticated user. Users can only access audiences they have created.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[GetAudienceResponse]
            Successfully retrieved audiences
        """
        _response = self._client_wrapper.httpx_client.request(
            "audience",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAudienceResponse,
                    construct_type(
                        type_=GetAudienceResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_audience_with_csv_upload(
        self,
        *,
        name: str,
        phone_number_column_name: str,
        identifier_column_name: str,
        file: core.File,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PostAudienceResponse]:
        """
        Create a new audience by uploading a CSV file containing phone numbers

        Parameters
        ----------
        name : str
            The name of the audience

        phone_number_column_name : str
            The name of the column in the CSV that contains phone numbers

        identifier_column_name : str
            The name of the column in the CSV that contains identifiers (e.g., names)

        file : core.File
            See core.File for more documentation

        description : typing.Optional[str]
            Optional description of the audience

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PostAudienceResponse]
            Audience created successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            "audience",
            base_url=self._client_wrapper.get_environment().atoms,
            method="POST",
            data={
                "name": name,
                "description": description,
                "phoneNumberColumnName": phone_number_column_name,
                "identifierColumnName": identifier_column_name,
            },
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
            force_multipart=True,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PostAudienceResponse,
                    construct_type(
                        type_=PostAudienceResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_audience_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[GetAudienceIdResponse]:
        """
        Retrieve a specific audience by its ID. Users can only access audiences they have created.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[GetAudienceIdResponse]
            Successfully retrieved audience
        """
        _response = self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAudienceIdResponse,
                    construct_type(
                        type_=GetAudienceIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_audience(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DeleteAudienceIdResponse]:
        """
        Delete a specific audience by its ID. Users can only delete audiences they created. Cannot delete if audience is being used by a campaign.

        Parameters
        ----------
        id : str
            The unique identifier of the audience to delete

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeleteAudienceIdResponse]
            Audience deleted successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteAudienceIdResponse,
                    construct_type(
                        type_=DeleteAudienceIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_audience_members(
        self,
        id: str,
        *,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[GetAudienceIdMembersResponse]:
        """
        Retrieve members of a specific audience with pagination support. Users can only access members of audiences they created.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        page : typing.Optional[int]
            Page number for pagination (default is 1)

        limit : typing.Optional[int]
            Number of items per page (default is 5)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[GetAudienceIdMembersResponse]
            Successfully retrieved audience members
        """
        _response = self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}/members",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            params={
                "page": page,
                "limit": limit,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAudienceIdMembersResponse,
                    construct_type(
                        type_=GetAudienceIdMembersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def add_audience_members(
        self,
        id: str,
        *,
        members: typing.Sequence[typing.Dict[str, typing.Any]],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PostAudienceIdMembersResponse]:
        """
        Add new members to an existing audience. Users can only add members to audiences they created.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        members : typing.Sequence[typing.Dict[str, typing.Any]]
            Array of member objects with dynamic structure based on audience configuration

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PostAudienceIdMembersResponse]
            Members added successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}/members",
            base_url=self._client_wrapper.get_environment().atoms,
            method="POST",
            json={
                "members": members,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PostAudienceIdMembersResponse,
                    construct_type(
                        type_=PostAudienceIdMembersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_audience_members(
        self, id: str, *, member_ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DeleteAudienceIdMembersResponse]:
        """
        Remove specific members from an audience by their member IDs. Users can only delete members from audiences they created.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        member_ids : typing.Sequence[str]
            Array of member IDs to delete

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeleteAudienceIdMembersResponse]
            Members deleted successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}/members",
            base_url=self._client_wrapper.get_environment().atoms,
            method="DELETE",
            json={
                "memberIds": member_ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteAudienceIdMembersResponse,
                    construct_type(
                        type_=DeleteAudienceIdMembersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def search_audience_members(
        self, id: str, *, query: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[GetAudienceIdMembersSearchResponse]:
        """
        Search for members within a specific audience using flexible search parameters. Users can only search members of audiences they created.

        **Search Types:**
        - **General Search** (`query`): Searches across all fields in the audience member data
        - **Field-Specific Search**: Use any field name as a parameter (e.g., `firstName=john`, `phoneNumber=123456`, `email=test@example.com`)

        **Examples:**
        - `?query=john` - General search across all fields
        - `?firstName=john` - Search specifically in firstName field
        - `?phoneNumber=555-1234` - Search specifically in phoneNumber field
        - `?firstName=john&lastName=doe` - Search for members matching both criteria

        **Note:** When using phoneNumber field, do not use quotes around the phone number. You can use either a general search OR field-specific searches, but not both simultaneously.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        query : typing.Optional[str]
            General search term that searches across all fields in audience member data

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[GetAudienceIdMembersSearchResponse]
            Search results returned successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}/members/search",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            params={
                "query": query,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAudienceIdMembersSearchResponse,
                    construct_type(
                        type_=GetAudienceIdMembersSearchResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawAudienceClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def get_all_audiences(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[GetAudienceResponse]:
        """
        Retrieve all audiences created by the authenticated user. Users can only access audiences they have created.

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[GetAudienceResponse]
            Successfully retrieved audiences
        """
        _response = await self._client_wrapper.httpx_client.request(
            "audience",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAudienceResponse,
                    construct_type(
                        type_=GetAudienceResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_audience_with_csv_upload(
        self,
        *,
        name: str,
        phone_number_column_name: str,
        identifier_column_name: str,
        file: core.File,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PostAudienceResponse]:
        """
        Create a new audience by uploading a CSV file containing phone numbers

        Parameters
        ----------
        name : str
            The name of the audience

        phone_number_column_name : str
            The name of the column in the CSV that contains phone numbers

        identifier_column_name : str
            The name of the column in the CSV that contains identifiers (e.g., names)

        file : core.File
            See core.File for more documentation

        description : typing.Optional[str]
            Optional description of the audience

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PostAudienceResponse]
            Audience created successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            "audience",
            base_url=self._client_wrapper.get_environment().atoms,
            method="POST",
            data={
                "name": name,
                "description": description,
                "phoneNumberColumnName": phone_number_column_name,
                "identifierColumnName": identifier_column_name,
            },
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
            force_multipart=True,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PostAudienceResponse,
                    construct_type(
                        type_=PostAudienceResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_audience_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[GetAudienceIdResponse]:
        """
        Retrieve a specific audience by its ID. Users can only access audiences they have created.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[GetAudienceIdResponse]
            Successfully retrieved audience
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAudienceIdResponse,
                    construct_type(
                        type_=GetAudienceIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_audience(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DeleteAudienceIdResponse]:
        """
        Delete a specific audience by its ID. Users can only delete audiences they created. Cannot delete if audience is being used by a campaign.

        Parameters
        ----------
        id : str
            The unique identifier of the audience to delete

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeleteAudienceIdResponse]
            Audience deleted successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteAudienceIdResponse,
                    construct_type(
                        type_=DeleteAudienceIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_audience_members(
        self,
        id: str,
        *,
        page: typing.Optional[int] = None,
        limit: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[GetAudienceIdMembersResponse]:
        """
        Retrieve members of a specific audience with pagination support. Users can only access members of audiences they created.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        page : typing.Optional[int]
            Page number for pagination (default is 1)

        limit : typing.Optional[int]
            Number of items per page (default is 5)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[GetAudienceIdMembersResponse]
            Successfully retrieved audience members
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}/members",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            params={
                "page": page,
                "limit": limit,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAudienceIdMembersResponse,
                    construct_type(
                        type_=GetAudienceIdMembersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def add_audience_members(
        self,
        id: str,
        *,
        members: typing.Sequence[typing.Dict[str, typing.Any]],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PostAudienceIdMembersResponse]:
        """
        Add new members to an existing audience. Users can only add members to audiences they created.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        members : typing.Sequence[typing.Dict[str, typing.Any]]
            Array of member objects with dynamic structure based on audience configuration

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PostAudienceIdMembersResponse]
            Members added successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}/members",
            base_url=self._client_wrapper.get_environment().atoms,
            method="POST",
            json={
                "members": members,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PostAudienceIdMembersResponse,
                    construct_type(
                        type_=PostAudienceIdMembersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_audience_members(
        self, id: str, *, member_ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DeleteAudienceIdMembersResponse]:
        """
        Remove specific members from an audience by their member IDs. Users can only delete members from audiences they created.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        member_ids : typing.Sequence[str]
            Array of member IDs to delete

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeleteAudienceIdMembersResponse]
            Members deleted successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}/members",
            base_url=self._client_wrapper.get_environment().atoms,
            method="DELETE",
            json={
                "memberIds": member_ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteAudienceIdMembersResponse,
                    construct_type(
                        type_=DeleteAudienceIdMembersResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def search_audience_members(
        self, id: str, *, query: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[GetAudienceIdMembersSearchResponse]:
        """
        Search for members within a specific audience using flexible search parameters. Users can only search members of audiences they created.

        **Search Types:**
        - **General Search** (`query`): Searches across all fields in the audience member data
        - **Field-Specific Search**: Use any field name as a parameter (e.g., `firstName=john`, `phoneNumber=123456`, `email=test@example.com`)

        **Examples:**
        - `?query=john` - General search across all fields
        - `?firstName=john` - Search specifically in firstName field
        - `?phoneNumber=555-1234` - Search specifically in phoneNumber field
        - `?firstName=john&lastName=doe` - Search for members matching both criteria

        **Note:** When using phoneNumber field, do not use quotes around the phone number. You can use either a general search OR field-specific searches, but not both simultaneously.

        Parameters
        ----------
        id : str
            The unique identifier of the audience

        query : typing.Optional[str]
            General search term that searches across all fields in audience member data

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[GetAudienceIdMembersSearchResponse]
            Search results returned successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"audience/{jsonable_encoder(id)}/members/search",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            params={
                "query": query,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAudienceIdMembersSearchResponse,
                    construct_type(
                        type_=GetAudienceIdMembersSearchResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ...core.pydantic_utilities import IS_PYDANTIC_V2
from ...core.serialization import FieldMetadata
from ...core.unchecked_base_model import UncheckedBaseModel
from .create_agent_request_background_sound import CreateAgentRequestBackgroundSound
from .create_agent_request_language import CreateAgentRequestLanguage
from .create_agent_request_slm_model import CreateAgentRequestSlmModel
from .create_agent_request_synthesizer import CreateAgentRequestSynthesizer
from .workflow_type import WorkflowType


class CreateAgentRequest(UncheckedBaseModel):
    name: str
    description: typing.Optional[str] = None
    background_sound: typing_extensions.Annotated[
        typing.Optional[CreateAgentRequestBackgroundSound],
        FieldMetadata(alias="backgroundSound"),
        pydantic.Field(
            alias="backgroundSound",
            description="Ambient background sound during calls. Options: '' (none), 'office', 'cafe', 'call_center', 'static'.",
        ),
    ] = None
    language: typing.Optional[CreateAgentRequestLanguage] = pydantic.Field(default=None)
    """
    Language configuration for the agent. You can enable or disable language switching for the agent. This will be used to determine the language of the agent.
    """

    synthesizer: typing.Optional[CreateAgentRequestSynthesizer] = pydantic.Field(default=None)
    """
    Synthesizer configuration for the agent. You can configure the synthesizer to use different voices and models. Currently we support 3 types of models for the synthesizer. Waves, Waves Lightning Large and Waves Lightning Large Voice Clone. You can clone your voice using waves platform https://waves.smallest.ai/voice-clone and use the voiceId for this field and select the model as waves_lightning_large_voice_clone to use your cloned voice. When updating the synthesizer configuration to voice clone model, you have to provide model and voiceId and gender all are required fields but when selecting the model as waves or waves and waves_lightning_large, you have to provide only model field and voiceId.
    """

    global_knowledge_base_id: typing_extensions.Annotated[
        typing.Optional[str],
        FieldMetadata(alias="globalKnowledgeBaseId"),
        pydantic.Field(
            alias="globalKnowledgeBaseId",
            description="The global knowledge base ID of the agent. You can create a global knowledge base by using the /knowledgebase endpoint and assign it to the agent. The agent will use this knowledge base for its responses.",
        ),
    ] = None
    slm_model: typing_extensions.Annotated[
        typing.Optional[CreateAgentRequestSlmModel],
        FieldMetadata(alias="slmModel"),
        pydantic.Field(
            alias="slmModel",
            description="The LLM model to use for the agent. LLM model will be used to generate the response and take decisions based on the user's query.",
        ),
    ] = None
    default_variables: typing_extensions.Annotated[
        typing.Optional[typing.Dict[str, typing.Any]],
        FieldMetadata(alias="defaultVariables"),
        pydantic.Field(
            alias="defaultVariables",
            description="The default variables to use for the agent. These variables will be used if no variables are provided when initiating a conversation with the agent.",
        ),
    ] = None
    global_prompt: typing_extensions.Annotated[
        typing.Optional[str],
        FieldMetadata(alias="globalPrompt"),
        pydantic.Field(
            alias="globalPrompt",
            description="Set global instructions for your agent's personality, role, and behavior throughout conversations. \nNote: Only used for workflow_graph agents.",
        ),
    ] = None
    telephony_product_id: typing_extensions.Annotated[
        typing.Optional[str],
        FieldMetadata(alias="telephonyProductId"),
        pydantic.Field(
            alias="telephonyProductId",
            description="The telephony product ID of the agent. This is the product ID of the telephony product that will be used to make the outbound call. You can buy telephone number and assign it to the agent.",
        ),
    ] = None
    workflow_type: typing_extensions.Annotated[
        typing.Optional[WorkflowType],
        FieldMetadata(alias="workflowType"),
        pydantic.Field(
            alias="workflowType",
            description="The type of workflow to create for the agent. Defaults to workflow_graph if not specified.",
        ),
    ] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
import urllib.parse
from contextlib import asynccontextmanager, contextmanager

import websockets.sync.client as websockets_sync_client
from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.jsonable_encoder import jsonable_encoder
from ...core.query_encoder import encode_query
from ...core.remove_none_from_dict import remove_none_from_dict
from ...core.request_options import RequestOptions
from ...core.websocket_compat import InvalidWebSocketStatus, get_status_code
from .raw_client import AsyncRawStreamingTtsClient, RawStreamingTtsClient
from .socket_client import AsyncStreamingTtsSocketClient, StreamingTtsSocketClient

try:
    from websockets.legacy.client import connect as websockets_client_connect  # type: ignore
except ImportError:
    from websockets import connect as websockets_client_connect  # type: ignore


class StreamingTtsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawStreamingTtsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawStreamingTtsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawStreamingTtsClient
        """
        return self._raw_client

    @contextmanager
    def connect(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.Iterator[StreamingTtsSocketClient]:
        """
        The Lightning SSE API allows you to stream text-to-speech audio in real-time. This is particularly useful for applications requiring low-latency audio generation.

        ### Connection

        Connect to the SSE endpoint:

        ```javascript
        const eventSource = new EventSource(
          'https://api.smallest.ai/api/v1/lightning/stream',
          {
            headers: {
              Authorization: 'Bearer YOUR_API_KEY'
            }
          }
        );
        ```

        ### Request Format

        Send a POST request with the following JSON structure:

        ```json
        {
          "text": "Your text to be converted to speech",
          "voice_id": "voice_id_here",
          "speed": 1,
          "sample_rate": 24000
        }
        ```

        #### Parameters

        | Parameter     | Type   | Required | Description                                         |
        | ------------- | ------ | -------- | --------------------------------------------------- |
        | `text`        | string | Yes      | The text to convert to speech (max 1000 characters) |
        | `voice_id`    | string | Yes      | ID of the voice to use                              |
        | `speed`       | number | No       | Speech speed multiplier (default: 1)                |
        | `sample_rate` | number | No       | Audio sample rate in Hz (default: 24000)            |

        ### Example Usage

        ```javascript
        const eventSource = new EventSource(
          'https://api.smallest.ai/api/v1/lightning/stream',
          {
            headers: {
              Authorization: 'Bearer YOUR_API_KEY'
            }
          }
        );

        // Handle incoming audio chunks
        eventSource.onmessage = (event) => {
          const response = JSON.parse(event.data);

          if (response.status === 'chunk') {
            // Decode and play audio
            const audioData = atob(response.data.audio);
            // Process audio data...
          } else if (response.status === 'complete' && response.done) {
            console.log('All audio chunks received');
            eventSource.close();
          } else if (response.status === 'error') {
            console.error('Error:', response.message);
            eventSource.close();
          }
        };

        eventSource.onerror = (error) => {
          console.error('SSE error:', error);
          eventSource.close();
        };
        ```

        ### Response Events

        The server will send events with the following formats:

        #### Chunk Event

        ```json
        {
          "request_id": "047c9091-b770-41d8-b96b-907d1c8406c0",
          "status": "chunk",
          "data": {
            "audio": "base64_encoded_audio_data"
          }
        }
        ```

        #### Complete Event

        ```json
        {
          "request_id": "047c9091-b770-41d8-b96b-907d1c8406c0",
          "status": "complete",
          "message": "All chunks sent",
          "done": true
        }
        ```

        #### Error Event

        ```json
        {
          "status": "error",
          "message": "Error message",
          "errors": [
            /* detailed error information * /
          ]
        }
        ```

        ### Notes

        - The API automatically chunks long text and streams each chunk separately
        - Credits are deducted based on the length of the input text
        - The SSE connection will remain open until all chunks are sent or an error occurs
        - For optimal performance, keep individual requests under 1000 characters

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        StreamingTtsSocketClient
        """
        ws_url = self._raw_client._client_wrapper.get_environment().waves_ws + "/api/v1/streaming-tts/stream"
        _encoded_query_params = encode_query(
            jsonable_encoder(
                remove_none_from_dict(
                    {
                        **(
                            request_options.get("additional_query_parameters", {}) or {}
                            if request_options is not None
                            else {}
                        ),
                    }
                )
            )
        )
        if _encoded_query_params:
            ws_url = ws_url + "?" + urllib.parse.urlencode(_encoded_query_params)
        headers = self._raw_client._client_wrapper.get_headers()
        if request_options and "additional_headers" in request_options:
            headers.update(request_options["additional_headers"])
        try:
            with websockets_sync_client.connect(ws_url, additional_headers=headers) as protocol:
                yield StreamingTtsSocketClient(websocket=protocol)
        except InvalidWebSocketStatus as exc:
            status_code: int = get_status_code(exc)
            if status_code == 401:
                raise ApiError(
                    status_code=status_code,
                    headers=dict(headers),
                    body="Websocket initialized with invalid credentials.",
                )
            raise ApiError(
                status_code=status_code,
                headers=dict(headers),
                body="Unexpected error when initializing websocket connection.",
            )


class AsyncStreamingTtsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawStreamingTtsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawStreamingTtsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawStreamingTtsClient
        """
        return self._raw_client

    @asynccontextmanager
    async def connect(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.AsyncIterator[AsyncStreamingTtsSocketClient]:
        """
        The Lightning SSE API allows you to stream text-to-speech audio in real-time. This is particularly useful for applications requiring low-latency audio generation.

        ### Connection

        Connect to the SSE endpoint:

        ```javascript
        const eventSource = new EventSource(
          'https://api.smallest.ai/api/v1/lightning/stream',
          {
            headers: {
              Authorization: 'Bearer YOUR_API_KEY'
            }
          }
        );
        ```

        ### Request Format

        Send a POST request with the following JSON structure:

        ```json
        {
          "text": "Your text to be converted to speech",
          "voice_id": "voice_id_here",
          "speed": 1,
          "sample_rate": 24000
        }
        ```

        #### Parameters

        | Parameter     | Type   | Required | Description                                         |
        | ------------- | ------ | -------- | --------------------------------------------------- |
        | `text`        | string | Yes      | The text to convert to speech (max 1000 characters) |
        | `voice_id`    | string | Yes      | ID of the voice to use                              |
        | `speed`       | number | No       | Speech speed multiplier (default: 1)                |
        | `sample_rate` | number | No       | Audio sample rate in Hz (default: 24000)            |

        ### Example Usage

        ```javascript
        const eventSource = new EventSource(
          'https://api.smallest.ai/api/v1/lightning/stream',
          {
            headers: {
              Authorization: 'Bearer YOUR_API_KEY'
            }
          }
        );

        // Handle incoming audio chunks
        eventSource.onmessage = (event) => {
          const response = JSON.parse(event.data);

          if (response.status === 'chunk') {
            // Decode and play audio
            const audioData = atob(response.data.audio);
            // Process audio data...
          } else if (response.status === 'complete' && response.done) {
            console.log('All audio chunks received');
            eventSource.close();
          } else if (response.status === 'error') {
            console.error('Error:', response.message);
            eventSource.close();
          }
        };

        eventSource.onerror = (error) => {
          console.error('SSE error:', error);
          eventSource.close();
        };
        ```

        ### Response Events

        The server will send events with the following formats:

        #### Chunk Event

        ```json
        {
          "request_id": "047c9091-b770-41d8-b96b-907d1c8406c0",
          "status": "chunk",
          "data": {
            "audio": "base64_encoded_audio_data"
          }
        }
        ```

        #### Complete Event

        ```json
        {
          "request_id": "047c9091-b770-41d8-b96b-907d1c8406c0",
          "status": "complete",
          "message": "All chunks sent",
          "done": true
        }
        ```

        #### Error Event

        ```json
        {
          "status": "error",
          "message": "Error message",
          "errors": [
            /* detailed error information * /
          ]
        }
        ```

        ### Notes

        - The API automatically chunks long text and streams each chunk separately
        - Credits are deducted based on the length of the input text
        - The SSE connection will remain open until all chunks are sent or an error occurs
        - For optimal performance, keep individual requests under 1000 characters

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncStreamingTtsSocketClient
        """
        ws_url = self._raw_client._client_wrapper.get_environment().waves_ws + "/api/v1/streaming-tts/stream"
        _encoded_query_params = encode_query(
            jsonable_encoder(
                remove_none_from_dict(
                    {
                        **(
                            request_options.get("additional_query_parameters", {}) or {}
                            if request_options is not None
                            else {}
                        ),
                    }
                )
            )
        )
        if _encoded_query_params:
            ws_url = ws_url + "?" + urllib.parse.urlencode(_encoded_query_params)
        headers = self._raw_client._client_wrapper.get_headers()
        if request_options and "additional_headers" in request_options:
            headers.update(request_options["additional_headers"])
        try:
            async with websockets_client_connect(ws_url, extra_headers=headers) as protocol:
                yield AsyncStreamingTtsSocketClient(websocket=protocol)
        except InvalidWebSocketStatus as exc:
            status_code: int = get_status_code(exc)
            if status_code == 401:
                raise ApiError(
                    status_code=status_code,
                    headers=dict(headers),
                    body="Websocket initialized with invalid credentials.",
                )
            raise ApiError(
                status_code=status_code,
                headers=dict(headers),
                body="Unexpected error when initializing websocket connection.",
            )



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.http_response import AsyncHttpResponse, HttpResponse
from ...core.jsonable_encoder import jsonable_encoder
from ...core.request_options import RequestOptions
from ...core.serialization import convert_and_respect_annotation_metadata
from ...core.unchecked_base_model import construct_type
from ..errors.bad_request_error import BadRequestError
from ..errors.forbidden_error import ForbiddenError
from ..errors.internal_server_error import InternalServerError
from ..errors.not_found_error import NotFoundError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.create_agent_request_background_sound import CreateAgentRequestBackgroundSound
from ..types.create_agent_request_language import CreateAgentRequestLanguage
from ..types.create_agent_request_slm_model import CreateAgentRequestSlmModel
from ..types.create_agent_request_synthesizer import CreateAgentRequestSynthesizer
from ..types.workflow_type import WorkflowType
from .types.delete_agent_agent_id_webhook_subscriptions_response import DeleteAgentAgentIdWebhookSubscriptionsResponse
from .types.delete_agent_id_response import DeleteAgentIdResponse
from .types.get_agent_agent_id_webhook_subscriptions_response import GetAgentAgentIdWebhookSubscriptionsResponse
from .types.get_agent_id_response import GetAgentIdResponse
from .types.get_agent_id_workflow_response import GetAgentIdWorkflowResponse
from .types.get_agent_response import GetAgentResponse
from .types.patch_agent_id_request_language import PatchAgentIdRequestLanguage
from .types.patch_agent_id_request_slm_model import PatchAgentIdRequestSlmModel
from .types.patch_agent_id_request_synthesizer import PatchAgentIdRequestSynthesizer
from .types.patch_agent_id_response import PatchAgentIdResponse
from .types.post_agent_agent_id_webhook_subscriptions_request_event_types_item import (
    PostAgentAgentIdWebhookSubscriptionsRequestEventTypesItem,
)
from .types.post_agent_agent_id_webhook_subscriptions_response import PostAgentAgentIdWebhookSubscriptionsResponse
from .types.post_agent_response import PostAgentResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawAgentsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def get_all_agents(
        self,
        *,
        page: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        search: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[GetAgentResponse]:
        """
        Agents are the main entities in the system. Agents are used to create conversations. You can create workflow for an agent and configure it for different use cases. You can also create custom workflows for an agent. This API will give you the list of agents created by organization you are a part of.

        Parameters
        ----------
        page : typing.Optional[int]

        offset : typing.Optional[int]

        search : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[GetAgentResponse]
            Successful response
        """
        _response = self._client_wrapper.httpx_client.request(
            "agent",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            params={
                "page": page,
                "offset": offset,
                "search": search,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAgentResponse,
                    construct_type(
                        type_=GetAgentResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_a_new_agent(
        self,
        *,
        name: str,
        description: typing.Optional[str] = OMIT,
        background_sound: typing.Optional[CreateAgentRequestBackgroundSound] = OMIT,
        language: typing.Optional[CreateAgentRequestLanguage] = OMIT,
        synthesizer: typing.Optional[CreateAgentRequestSynthesizer] = OMIT,
        global_knowledge_base_id: typing.Optional[str] = OMIT,
        slm_model: typing.Optional[CreateAgentRequestSlmModel] = OMIT,
        default_variables: typing.Optional[typing.Dict[str, typing.Any]] = OMIT,
        global_prompt: typing.Optional[str] = OMIT,
        telephony_product_id: typing.Optional[str] = OMIT,
        workflow_type: typing.Optional[WorkflowType] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PostAgentResponse]:
        """
        You can create a new agent by passing the name of the agent in the request body. You can use update-workflow endpoint next to assign custom workflow to the agent.

        Parameters
        ----------
        name : str

        description : typing.Optional[str]

        background_sound : typing.Optional[CreateAgentRequestBackgroundSound]
            Ambient background sound during calls. Options: '' (none), 'office', 'cafe', 'call_center', 'static'.

        language : typing.Optional[CreateAgentRequestLanguage]
            Language configuration for the agent. You can enable or disable language switching for the agent. This will be used to determine the language of the agent.

        synthesizer : typing.Optional[CreateAgentRequestSynthesizer]
            Synthesizer configuration for the agent. You can configure the synthesizer to use different voices and models. Currently we support 3 types of models for the synthesizer. Waves, Waves Lightning Large and Waves Lightning Large Voice Clone. You can clone your voice using waves platform https://waves.smallest.ai/voice-clone and use the voiceId for this field and select the model as waves_lightning_large_voice_clone to use your cloned voice. When updating the synthesizer configuration to voice clone model, you have to provide model and voiceId and gender all are required fields but when selecting the model as waves or waves and waves_lightning_large, you have to provide only model field and voiceId.

        global_knowledge_base_id : typing.Optional[str]
            The global knowledge base ID of the agent. You can create a global knowledge base by using the /knowledgebase endpoint and assign it to the agent. The agent will use this knowledge base for its responses.

        slm_model : typing.Optional[CreateAgentRequestSlmModel]
            The LLM model to use for the agent. LLM model will be used to generate the response and take decisions based on the user's query.

        default_variables : typing.Optional[typing.Dict[str, typing.Any]]
            The default variables to use for the agent. These variables will be used if no variables are provided when initiating a conversation with the agent.

        global_prompt : typing.Optional[str]
            Set global instructions for your agent's personality, role, and behavior throughout conversations.
            Note: Only used for workflow_graph agents.

        telephony_product_id : typing.Optional[str]
            The telephony product ID of the agent. This is the product ID of the telephony product that will be used to make the outbound call. You can buy telephone number and assign it to the agent.

        workflow_type : typing.Optional[WorkflowType]
            The type of workflow to create for the agent. Defaults to workflow_graph if not specified.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PostAgentResponse]
            Agent created successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            "agent",
            base_url=self._client_wrapper.get_environment().atoms,
            method="POST",
            json={
                "name": name,
                "description": description,
                "backgroundSound": background_sound,
                "language": convert_and_respect_annotation_metadata(
                    object_=language, annotation=CreateAgentRequestLanguage, direction="write"
                ),
                "synthesizer": convert_and_respect_annotation_metadata(
                    object_=synthesizer, annotation=CreateAgentRequestSynthesizer, direction="write"
                ),
                "globalKnowledgeBaseId": global_knowledge_base_id,
                "slmModel": slm_model,
                "defaultVariables": default_variables,
                "globalPrompt": global_prompt,
                "telephonyProductId": telephony_product_id,
                "workflowType": workflow_type,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PostAgentResponse,
                    construct_type(
                        type_=PostAgentResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_agent_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[GetAgentIdResponse]:
        """
        Agents are the main entities in the system. Agents are used to create conversations. You can create workflow for an agent and configure it for different use cases. You can also create custom workflows for an agent.

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[GetAgentIdResponse]
            Successful response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAgentIdResponse,
                    construct_type(
                        type_=GetAgentIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_an_agent(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DeleteAgentIdResponse]:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeleteAgentIdResponse]
            Agent deleted successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteAgentIdResponse,
                    construct_type(
                        type_=DeleteAgentIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update_an_agent(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        language: typing.Optional[PatchAgentIdRequestLanguage] = OMIT,
        synthesizer: typing.Optional[PatchAgentIdRequestSynthesizer] = OMIT,
        global_knowledge_base_id: typing.Optional[str] = OMIT,
        slm_model: typing.Optional[PatchAgentIdRequestSlmModel] = OMIT,
        default_variables: typing.Optional[typing.Dict[str, typing.Any]] = OMIT,
        global_prompt: typing.Optional[str] = OMIT,
        telephony_product_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PatchAgentIdResponse]:
        """
        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        description : typing.Optional[str]

        language : typing.Optional[PatchAgentIdRequestLanguage]
            Language configuration for the agent. You can enable or disable language switching for the agent. This will be used to determine the language of the agent.

        synthesizer : typing.Optional[PatchAgentIdRequestSynthesizer]
            Synthesizer configuration for the agent. You can configure the synthesizer to use different voices and models. Currently we support 3 types of models for the synthesizer. Waves, Waves Lightning Large and Waves Lightning Large Voice Clone. You can clone your voice using waves platform https://waves.smallest.ai/voice-clone and use the voiceId for this field and select the model as waves_lightning_large_voice_clone to use your cloned voice. When updating the synthesizer configuration to voice clone model, you have to provide model and voiceId and gender all are required fields but when selecting the model as waves or waves and waves_lightning_large, you have to provide only model field and voiceId.

        global_knowledge_base_id : typing.Optional[str]

        slm_model : typing.Optional[PatchAgentIdRequestSlmModel]

        default_variables : typing.Optional[typing.Dict[str, typing.Any]]
            The default variables to use for the agent. These variables will be used if no variables are provided when initiating a conversation with the agent.

        global_prompt : typing.Optional[str]
            Set global instructions for your agent's personality, role, and behavior throughout conversations.
            Note: Only used for workflow_graph agents.

        telephony_product_id : typing.Optional[str]
            The telephony product ID of the agent. This is the product ID of the telephony product that will be used to make the outbound call. You can buy telephone number and assign it to the agent.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PatchAgentIdResponse]
            Agent updated successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="PATCH",
            json={
                "name": name,
                "description": description,
                "language": convert_and_respect_annotation_metadata(
                    object_=language, annotation=PatchAgentIdRequestLanguage, direction="write"
                ),
                "synthesizer": convert_and_respect_annotation_metadata(
                    object_=synthesizer, annotation=PatchAgentIdRequestSynthesizer, direction="write"
                ),
                "globalKnowledgeBaseId": global_knowledge_base_id,
                "slmModel": slm_model,
                "defaultVariables": default_variables,
                "globalPrompt": global_prompt,
                "telephonyProductId": telephony_product_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PatchAgentIdResponse,
                    construct_type(
                        type_=PatchAgentIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_agent_workflow(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[GetAgentIdWorkflowResponse]:
        """
        Retrieve the workflow configuration for a specific agent. The workflow contains nodes and edges that define the conversation flow and logic for the agent.

        Parameters
        ----------
        id : str
            The ID of the agent

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[GetAgentIdWorkflowResponse]
            Successful response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(id)}/workflow",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAgentIdWorkflowResponse,
                    construct_type(
                        type_=GetAgentIdWorkflowResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_webhook_subscriptions_for_an_agent(
        self, agent_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[GetAgentAgentIdWebhookSubscriptionsResponse]:
        """
        Retrieve webhook subscriptions for a given agent ID

        Parameters
        ----------
        agent_id : str
            The ID of the agent

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[GetAgentAgentIdWebhookSubscriptionsResponse]
            Successful response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(agent_id)}/webhook-subscriptions",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAgentAgentIdWebhookSubscriptionsResponse,
                    construct_type(
                        type_=GetAgentAgentIdWebhookSubscriptionsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_webhook_subscriptions_for_an_agent(
        self,
        agent_id: str,
        *,
        event_types: typing.Sequence[PostAgentAgentIdWebhookSubscriptionsRequestEventTypesItem],
        webhook_id: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PostAgentAgentIdWebhookSubscriptionsResponse]:
        """
        Create webhook subscriptions for a specific agent with selected event types

        Parameters
        ----------
        agent_id : str
            The ID of the agent to create subscriptions for

        event_types : typing.Sequence[PostAgentAgentIdWebhookSubscriptionsRequestEventTypesItem]
            Array of event types to subscribe to

        webhook_id : str
            The ID of the webhook to subscribe to

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PostAgentAgentIdWebhookSubscriptionsResponse]
            Subscriptions created successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(agent_id)}/webhook-subscriptions",
            base_url=self._client_wrapper.get_environment().atoms,
            method="POST",
            json={
                "eventTypes": event_types,
                "webhookId": webhook_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PostAgentAgentIdWebhookSubscriptionsResponse,
                    construct_type(
                        type_=PostAgentAgentIdWebhookSubscriptionsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_webhook_subscriptions_for_an_agent(
        self, agent_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[DeleteAgentAgentIdWebhookSubscriptionsResponse]:
        """
        Delete webhook subscriptions for a given agent ID

        Parameters
        ----------
        agent_id : str
            The ID of the agent to filter subscriptions by

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[DeleteAgentAgentIdWebhookSubscriptionsResponse]
            Subscriptions deleted successfully
        """
        _response = self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(agent_id)}/webhook-subscriptions",
            base_url=self._client_wrapper.get_environment().atoms,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteAgentAgentIdWebhookSubscriptionsResponse,
                    construct_type(
                        type_=DeleteAgentAgentIdWebhookSubscriptionsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawAgentsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def get_all_agents(
        self,
        *,
        page: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        search: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[GetAgentResponse]:
        """
        Agents are the main entities in the system. Agents are used to create conversations. You can create workflow for an agent and configure it for different use cases. You can also create custom workflows for an agent. This API will give you the list of agents created by organization you are a part of.

        Parameters
        ----------
        page : typing.Optional[int]

        offset : typing.Optional[int]

        search : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[GetAgentResponse]
            Successful response
        """
        _response = await self._client_wrapper.httpx_client.request(
            "agent",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            params={
                "page": page,
                "offset": offset,
                "search": search,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAgentResponse,
                    construct_type(
                        type_=GetAgentResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_a_new_agent(
        self,
        *,
        name: str,
        description: typing.Optional[str] = OMIT,
        background_sound: typing.Optional[CreateAgentRequestBackgroundSound] = OMIT,
        language: typing.Optional[CreateAgentRequestLanguage] = OMIT,
        synthesizer: typing.Optional[CreateAgentRequestSynthesizer] = OMIT,
        global_knowledge_base_id: typing.Optional[str] = OMIT,
        slm_model: typing.Optional[CreateAgentRequestSlmModel] = OMIT,
        default_variables: typing.Optional[typing.Dict[str, typing.Any]] = OMIT,
        global_prompt: typing.Optional[str] = OMIT,
        telephony_product_id: typing.Optional[str] = OMIT,
        workflow_type: typing.Optional[WorkflowType] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PostAgentResponse]:
        """
        You can create a new agent by passing the name of the agent in the request body. You can use update-workflow endpoint next to assign custom workflow to the agent.

        Parameters
        ----------
        name : str

        description : typing.Optional[str]

        background_sound : typing.Optional[CreateAgentRequestBackgroundSound]
            Ambient background sound during calls. Options: '' (none), 'office', 'cafe', 'call_center', 'static'.

        language : typing.Optional[CreateAgentRequestLanguage]
            Language configuration for the agent. You can enable or disable language switching for the agent. This will be used to determine the language of the agent.

        synthesizer : typing.Optional[CreateAgentRequestSynthesizer]
            Synthesizer configuration for the agent. You can configure the synthesizer to use different voices and models. Currently we support 3 types of models for the synthesizer. Waves, Waves Lightning Large and Waves Lightning Large Voice Clone. You can clone your voice using waves platform https://waves.smallest.ai/voice-clone and use the voiceId for this field and select the model as waves_lightning_large_voice_clone to use your cloned voice. When updating the synthesizer configuration to voice clone model, you have to provide model and voiceId and gender all are required fields but when selecting the model as waves or waves and waves_lightning_large, you have to provide only model field and voiceId.

        global_knowledge_base_id : typing.Optional[str]
            The global knowledge base ID of the agent. You can create a global knowledge base by using the /knowledgebase endpoint and assign it to the agent. The agent will use this knowledge base for its responses.

        slm_model : typing.Optional[CreateAgentRequestSlmModel]
            The LLM model to use for the agent. LLM model will be used to generate the response and take decisions based on the user's query.

        default_variables : typing.Optional[typing.Dict[str, typing.Any]]
            The default variables to use for the agent. These variables will be used if no variables are provided when initiating a conversation with the agent.

        global_prompt : typing.Optional[str]
            Set global instructions for your agent's personality, role, and behavior throughout conversations.
            Note: Only used for workflow_graph agents.

        telephony_product_id : typing.Optional[str]
            The telephony product ID of the agent. This is the product ID of the telephony product that will be used to make the outbound call. You can buy telephone number and assign it to the agent.

        workflow_type : typing.Optional[WorkflowType]
            The type of workflow to create for the agent. Defaults to workflow_graph if not specified.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PostAgentResponse]
            Agent created successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            "agent",
            base_url=self._client_wrapper.get_environment().atoms,
            method="POST",
            json={
                "name": name,
                "description": description,
                "backgroundSound": background_sound,
                "language": convert_and_respect_annotation_metadata(
                    object_=language, annotation=CreateAgentRequestLanguage, direction="write"
                ),
                "synthesizer": convert_and_respect_annotation_metadata(
                    object_=synthesizer, annotation=CreateAgentRequestSynthesizer, direction="write"
                ),
                "globalKnowledgeBaseId": global_knowledge_base_id,
                "slmModel": slm_model,
                "defaultVariables": default_variables,
                "globalPrompt": global_prompt,
                "telephonyProductId": telephony_product_id,
                "workflowType": workflow_type,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PostAgentResponse,
                    construct_type(
                        type_=PostAgentResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_agent_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[GetAgentIdResponse]:
        """
        Agents are the main entities in the system. Agents are used to create conversations. You can create workflow for an agent and configure it for different use cases. You can also create custom workflows for an agent.

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[GetAgentIdResponse]
            Successful response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAgentIdResponse,
                    construct_type(
                        type_=GetAgentIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_an_agent(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DeleteAgentIdResponse]:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeleteAgentIdResponse]
            Agent deleted successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteAgentIdResponse,
                    construct_type(
                        type_=DeleteAgentIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update_an_agent(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        language: typing.Optional[PatchAgentIdRequestLanguage] = OMIT,
        synthesizer: typing.Optional[PatchAgentIdRequestSynthesizer] = OMIT,
        global_knowledge_base_id: typing.Optional[str] = OMIT,
        slm_model: typing.Optional[PatchAgentIdRequestSlmModel] = OMIT,
        default_variables: typing.Optional[typing.Dict[str, typing.Any]] = OMIT,
        global_prompt: typing.Optional[str] = OMIT,
        telephony_product_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PatchAgentIdResponse]:
        """
        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        description : typing.Optional[str]

        language : typing.Optional[PatchAgentIdRequestLanguage]
            Language configuration for the agent. You can enable or disable language switching for the agent. This will be used to determine the language of the agent.

        synthesizer : typing.Optional[PatchAgentIdRequestSynthesizer]
            Synthesizer configuration for the agent. You can configure the synthesizer to use different voices and models. Currently we support 3 types of models for the synthesizer. Waves, Waves Lightning Large and Waves Lightning Large Voice Clone. You can clone your voice using waves platform https://waves.smallest.ai/voice-clone and use the voiceId for this field and select the model as waves_lightning_large_voice_clone to use your cloned voice. When updating the synthesizer configuration to voice clone model, you have to provide model and voiceId and gender all are required fields but when selecting the model as waves or waves and waves_lightning_large, you have to provide only model field and voiceId.

        global_knowledge_base_id : typing.Optional[str]

        slm_model : typing.Optional[PatchAgentIdRequestSlmModel]

        default_variables : typing.Optional[typing.Dict[str, typing.Any]]
            The default variables to use for the agent. These variables will be used if no variables are provided when initiating a conversation with the agent.

        global_prompt : typing.Optional[str]
            Set global instructions for your agent's personality, role, and behavior throughout conversations.
            Note: Only used for workflow_graph agents.

        telephony_product_id : typing.Optional[str]
            The telephony product ID of the agent. This is the product ID of the telephony product that will be used to make the outbound call. You can buy telephone number and assign it to the agent.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PatchAgentIdResponse]
            Agent updated successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().atoms,
            method="PATCH",
            json={
                "name": name,
                "description": description,
                "language": convert_and_respect_annotation_metadata(
                    object_=language, annotation=PatchAgentIdRequestLanguage, direction="write"
                ),
                "synthesizer": convert_and_respect_annotation_metadata(
                    object_=synthesizer, annotation=PatchAgentIdRequestSynthesizer, direction="write"
                ),
                "globalKnowledgeBaseId": global_knowledge_base_id,
                "slmModel": slm_model,
                "defaultVariables": default_variables,
                "globalPrompt": global_prompt,
                "telephonyProductId": telephony_product_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PatchAgentIdResponse,
                    construct_type(
                        type_=PatchAgentIdResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_agent_workflow(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[GetAgentIdWorkflowResponse]:
        """
        Retrieve the workflow configuration for a specific agent. The workflow contains nodes and edges that define the conversation flow and logic for the agent.

        Parameters
        ----------
        id : str
            The ID of the agent

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[GetAgentIdWorkflowResponse]
            Successful response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(id)}/workflow",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAgentIdWorkflowResponse,
                    construct_type(
                        type_=GetAgentIdWorkflowResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_webhook_subscriptions_for_an_agent(
        self, agent_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[GetAgentAgentIdWebhookSubscriptionsResponse]:
        """
        Retrieve webhook subscriptions for a given agent ID

        Parameters
        ----------
        agent_id : str
            The ID of the agent

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[GetAgentAgentIdWebhookSubscriptionsResponse]
            Successful response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(agent_id)}/webhook-subscriptions",
            base_url=self._client_wrapper.get_environment().atoms,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    GetAgentAgentIdWebhookSubscriptionsResponse,
                    construct_type(
                        type_=GetAgentAgentIdWebhookSubscriptionsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_webhook_subscriptions_for_an_agent(
        self,
        agent_id: str,
        *,
        event_types: typing.Sequence[PostAgentAgentIdWebhookSubscriptionsRequestEventTypesItem],
        webhook_id: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PostAgentAgentIdWebhookSubscriptionsResponse]:
        """
        Create webhook subscriptions for a specific agent with selected event types

        Parameters
        ----------
        agent_id : str
            The ID of the agent to create subscriptions for

        event_types : typing.Sequence[PostAgentAgentIdWebhookSubscriptionsRequestEventTypesItem]
            Array of event types to subscribe to

        webhook_id : str
            The ID of the webhook to subscribe to

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PostAgentAgentIdWebhookSubscriptionsResponse]
            Subscriptions created successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(agent_id)}/webhook-subscriptions",
            base_url=self._client_wrapper.get_environment().atoms,
            method="POST",
            json={
                "eventTypes": event_types,
                "webhookId": webhook_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PostAgentAgentIdWebhookSubscriptionsResponse,
                    construct_type(
                        type_=PostAgentAgentIdWebhookSubscriptionsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_webhook_subscriptions_for_an_agent(
        self, agent_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[DeleteAgentAgentIdWebhookSubscriptionsResponse]:
        """
        Delete webhook subscriptions for a given agent ID

        Parameters
        ----------
        agent_id : str
            The ID of the agent to filter subscriptions by

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[DeleteAgentAgentIdWebhookSubscriptionsResponse]
            Subscriptions deleted successfully
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"agent/{jsonable_encoder(agent_id)}/webhook-subscriptions",
            base_url=self._client_wrapper.get_environment().atoms,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    DeleteAgentAgentIdWebhookSubscriptionsResponse,
                    construct_type(
                        type_=DeleteAgentAgentIdWebhookSubscriptionsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Any,
                        construct_type(
                            type_=typing.Any,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

import typing

from ...types.single_prompt_data import SinglePromptData
from ...types.workflow_graph_data import WorkflowGraphData

GetAgentIdWorkflowResponseDataData = typing.Union[WorkflowGraphData, SinglePromptData]



# This file was auto-generated by Fern from our API Definition.

import typing

from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.request_options import RequestOptions
from .raw_client import AsyncRawVoicesClient, RawVoicesClient
from .types.get_waves_voices_request_model import GetWavesVoicesRequestModel
from .types.get_waves_voices_response import GetWavesVoicesResponse


class VoicesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawVoicesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawVoicesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawVoicesClient
        """
        return self._raw_client

    def get_waves_voices(
        self, model: GetWavesVoicesRequestModel, *, request_options: typing.Optional[RequestOptions] = None
    ) -> GetWavesVoicesResponse:
        """
        Get voices supported for a given model using the new Waves API.

        Parameters
        ----------
        model : GetWavesVoicesRequestModel
            The model to use for speech synthesis.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        GetWavesVoicesResponse
            Voices retrieved successfully.

        Examples
        --------
        from smallestai import SmallestAI

        client = SmallestAI(
            token="YOUR_TOKEN",
        )
        client.waves.voices.get_waves_voices(
            model="lightning",
        )
        """
        _response = self._raw_client.get_waves_voices(model, request_options=request_options)
        return _response.data


class AsyncVoicesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawVoicesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawVoicesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawVoicesClient
        """
        return self._raw_client

    async def get_waves_voices(
        self, model: GetWavesVoicesRequestModel, *, request_options: typing.Optional[RequestOptions] = None
    ) -> GetWavesVoicesResponse:
        """
        Get voices supported for a given model using the new Waves API.

        Parameters
        ----------
        model : GetWavesVoicesRequestModel
            The model to use for speech synthesis.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        GetWavesVoicesResponse
            Voices retrieved successfully.

        Examples
        --------
        import asyncio

        from smallestai import AsyncSmallestAI

        client = AsyncSmallestAI(
            token="YOUR_TOKEN",
        )


        async def main() -> None:
            await client.waves.voices.get_waves_voices(
                model="lightning",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.get_waves_voices(model, request_options=request_options)
        return _response.data



"""
Graph - Tree/DAG structure for agent nodes.

Supports building graphs from edges with validation:
- Multiple roots allowed
- No cycles
- Automatic sink connection for leaves
"""

import asyncio
import inspect
import json
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Dict, List, Optional, Set, Tuple

from fastapi import WebSocket, WebSocketDisconnect
from loguru import logger
from pydantic import BaseModel

from smallestai.atoms.agent.events import (
    SDKAgentErrorEvent,
    SDKAgentReadyEvent,
    SDKEvent,
    SDKSystemInitEvent,
)
from smallestai.atoms.agent.nodes import Node
from smallestai.atoms.agent.task_manager import TaskManager, TaskManagerParams


@dataclass
class EventHandler:
    name: str
    handlers: List[Any]


class SessionConfig(BaseModel):
    """Configuration for the session"""

    edges: List[Tuple[Node, Node]] = []
    nodes: List[Node] = []
    name: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True


class EventCodec:
    def encode(self, data: SDKEvent) -> bytes:
        return json.dumps(
            {
                "type": data.type,
                **data.model_dump(mode="json"),
            }
        ).encode()

    def decode(self, data: Dict[str, Any]) -> SDKEvent:
        return SDKEvent.parse_obj(data)


class RootNode(Node):
    """
    Root node - entry point that receives events from WebSocket.
    """

    def __init__(self, name: str = "Root"):
        super().__init__(name=name)

    async def process_event(self, event: SDKEvent):
        await self.send_event(event)


class SinkNode(Node):
    """
    Sink node - exit point that sends events to WebSocket.

    All leaf nodes in the graph connect to this sink.
    """

    def __init__(self, session: "AgentSession", name: str = "Sink"):
        super().__init__(name=name)
        self._session = session

    async def process_event(self, event: SDKEvent):
        """Send event to WebSocket via session"""
        if self._session:
            await self._session.send_to_websocket(event)
        else:
            logger.warning(f"[{self.name}] No session available to send event")


class AgentSession:
    """
    Agent session that manages graph execution and WebSocket communication.

    Handles:
    - WebSocket connection lifecycle
    - Event routing from WebSocket to graph
    - Event routing from graph (sink) to WebSocket
    - Graph lifecycle management
    """

    def __init__(
        self,
        websocket: WebSocket,
        session_id: str,
        setup_handler: Callable[["AgentSession"], Awaitable[None]],
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """
        Initialize the agent session.

        Args:
            websocket: FastAPI WebSocket connection
            session_id: Unique session identifier
        """
        self.name = f"Session-{session_id}"

        self.websocket = websocket
        self.session_id = session_id

        self._setup_handler = setup_handler

        self.nodes: List[Node] = []
        self.edges: List[Tuple[Node, Node]] = []

        self.root = RootNode(name=f"{self.name}::Root")
        self.sink = SinkNode(session=self, name=f"{self.name}::Sink")

        self.codec = EventCodec()
        self._running = False

        self._init_event: Optional[SDKSystemInitEvent] = None

        self.task_manager = TaskManager()

        self.loop = loop or asyncio.get_event_loop()

        self._receive_loop_task: Optional[asyncio.Task] = None

        self._shutdown_event = asyncio.Event()
        self._cleanup_complete = asyncio.Event()

        self._event_tasks: Set[asyncio.Task] = set()
        self._event_handlers: Dict[str, EventHandler] = {}

        self._register_event_handler("on_event_received")

    async def initialize(self):
        """Initialize the session"""
        logger.info(f"[{self.name}] Initializing session")

        self.task_manager.setup(TaskManagerParams(loop=self.loop))

        await self.websocket.accept()

        await self._handle_init_event()

        await self._setup_handler(self)

    async def _handle_init_event(self) -> None:
        """
        Handle the handshake protocol.

        Returns:
            bool: True if handshake successful, False otherwise
        """
        try:
            message = await asyncio.wait_for(
                self.websocket.receive_json(mode="binary"), timeout=10.0
            )
            logger.info(f"Received message: {message}")

            init_event = self.codec.decode(message)
            logger.info(f"Received handshake event: {init_event}")

            if not isinstance(init_event, SDKSystemInitEvent):
                logger.error(f"Expected HandshakeEvent, got {type(init_event)}")
                error_event = SDKAgentErrorEvent(
                    message=f"Expected InitEvent, got {type(init_event)}"
                )
                await self.send_to_websocket(error_event)
                # TODO: End the pipeline means disconnect the websocket because we have to not called the run handler

            if isinstance(init_event, SDKSystemInitEvent):
                self._init_event = init_event
                logger.info(
                    f"Received handshake: version={init_event.version}, "
                    f"conversation_type={init_event.session_context.conversation_type}"
                )

        except asyncio.TimeoutError:
            logger.error("Init timeout")
            # TODO: End the pipeline means disconnect the websocket because we have to not called the run handler

        except Exception as e:
            logger.error(f"Init error: {e}")
            # TODO: End the pipeline means disconnect the websocket because we have to not called the run handler

    def add_node(self, node: Node):
        """Add a node to the session graph"""
        self.nodes.append(node)
        logger.debug(f"[{self.name}] Added node: {node.name}")

    def add_edge(self, parent: Node, child: Node):
        """Add an edge between nodes"""
        self.edges.append((parent, child))
        logger.debug(f"[{self.name}] Added edge: {parent.name} -> {child.name}")

    async def start(self) -> None:
        """Start the session"""
        logger.info(f"[{self.name}] Starting session")
        if not self._init_event:
            logger.error(
                "This should not happen because this method should always be called after the init event is received which will set the init event"
            )
            raise ValueError("Session not initialized")

        logger.info(f"[{self.name}] Building graph with {len(self.nodes)} nodes")
        self._build_graph()

        self._running = True

        await self._start_nodes(self._init_event, self.task_manager)
        await self.send_to_websocket(SDKAgentReadyEvent())

        self._receive_loop_task = self.task_manager.create_task(
            self._receive_loop(), name="receive_loop"
        )

    def _build_graph(self):
        """Build and validate the graph"""
        for parent, child in self.edges:
            parent.add_child(child)

        nodes_with_no_parents = [node for node in self.nodes if not node.has_parents]
        for node in nodes_with_no_parents:
            self.root.add_child(node)

        nodes_with_no_children = [node for node in self.nodes if not node.has_children]

        for node in nodes_with_no_children:
            self.sink.add_parent(node)

        self.nodes = [self.root] + self.nodes + [self.sink]

        logger.info(f"[{self.name}] Graph built: {len(self.nodes)} nodes, ")

        if self._has_cycle():
            raise ValueError(f"[{self.name}] Graph contains cycles!")

    def _has_cycle(self) -> bool:
        """Detect if graph has cycles using DFS with colors"""
        WHITE, GRAY, BLACK = 0, 1, 2
        color = {node: WHITE for node in self.nodes}

        def dfs(node: Node) -> bool:
            if color[node] == GRAY:
                return True
            if color[node] == BLACK:
                return False

            color[node] = GRAY

            for child in node.children:
                if child in color and dfs(child):
                    return True

            color[node] = BLACK
            return False

        if dfs(self.root):
            return True

        return False

    async def _start_nodes(
        self, init_event: SDKSystemInitEvent, task_manager: TaskManager
    ):
        """Start all nodes including sink"""
        for node in self.nodes:
            await node.start(init_event, task_manager)
        logger.info(f"[{self.name}] All nodes started")

    async def _receive_loop(self):
        """Receive and process messages from WebSocket"""
        try:
            while self._running:
                data = await self.websocket.receive_json(mode="binary")
                try:
                    event = self.codec.decode(data)
                    await self.root.queue_event(event)

                    await self._call_event_handler("on_event_received", event)

                except Exception as e:
                    logger.error(f"Error in receive loop: {e}", exc_info=True)
        except WebSocketDisconnect:
            logger.info(f"[{self.name}] WebSocket disconnected in receive loop")
            await self.cleanup()
        except Exception as e:
            logger.error(f"[{self.name}] Error in receive loop: {e}", exc_info=True)
            await self.cleanup()

    async def send_to_websocket(self, event: SDKEvent):
        """Send event to WebSocket (called by sink)"""
        try:
            encoded_event = self.codec.encode(event)
            await self.websocket.send_bytes(encoded_event)
            logger.trace(f"[{self.name}] Sent to WebSocket: {event.type}")
        except Exception as e:
            logger.error(f"[{self.name}] Failed to send to WebSocket: {e}")

    async def stop(self):
        """Stop the session"""
        for node in self.nodes:
            await node.stop()

    async def cleanup(self):
        """Cleanup session resources"""
        if not self._running:
            return

        logger.info(f"[{self.name}] Starting cleanup")
        self._running = False

        for task in self._event_tasks:
            if self.task_manager:
                await self.task_manager.cancel_task(task)
        self._event_tasks.clear()

        await self.stop()

        try:
            await self.websocket.close()
            logger.debug(f"[{self.name}] WebSocket closed in cleanup")
        except Exception:
            logger.debug("Websocket already closed in cleanup")

        if self._receive_loop_task:
            await self.task_manager.cancel_task(self._receive_loop_task)
            self._receive_loop_task = None

        current_tasks = self.task_manager.current_tasks()
        if current_tasks:
            task_names = ", ".join(list(self.task_manager._tasks.keys()))
            logger.info(f"[{self.name}] Tasks: {task_names}")
            logger.info(
                f"[{self.name}] Waiting for {len(current_tasks)} tasks to complete"
            )
            await asyncio.gather(*current_tasks, return_exceptions=True)

        self._cleanup_complete.set()
        logger.info(f"[{self.name}] Cleanup complete")

    async def wait_until_complete(self):
        """Wait until the session is complete and cleaned up."""
        await self._cleanup_complete.wait()
        logger.info(f"[{self.name}] Session complete")

    def on_event(self, event_name: str):
        """Decorator for registering event handlers.
        Args:
            event_name: The name of the event to handle.
        Returns:
            The decorator function that registers the handler.
        """

        def decorator(handler):
            self.add_event_handler(event_name, handler)
            return handler

        return decorator

    def add_event_handler(self, event_name: str, handler):
        """Add an event handler for the specified event.
        Args:
            event_name: The name of the event to handle.
            handler: The function to call when the event occurs.
                Can be sync or async.
        """
        if event_name in self._event_handlers:
            self._event_handlers[event_name].handlers.append(handler)
        else:
            logger.warning(f"Event handler {event_name} not registered")

    def _register_event_handler(self, event_name: str):
        """Register an event handler type.
        Args:
            event_name: The name of the event type to register.
            sync: Whether this event handler will be executed in a task.
        """
        if event_name not in self._event_handlers:
            self._event_handlers[event_name] = EventHandler(
                name=event_name, handlers=[]
            )
        else:
            logger.warning(f"Event handler {event_name} already registered")

    async def _run_handler(self, event_name: str, handler, *args, **kwargs):
        """Execute all handlers for an event.
        Args:
            event_name: The event name for this handler.
            handler: The handler function to run.
            *args: Positional arguments to pass to handlers.
            **kwargs: Keyword arguments to pass to handlers.
        """
        try:
            if inspect.iscoroutinefunction(handler):
                await handler(self, *args, **kwargs)
            else:
                handler(self, *args, **kwargs)
        except Exception as e:
            logger.exception(f"Exception in event handler {event_name}: {e}")

    async def _call_event_handler(self, event_name: str, *args, **kwargs):
        """Call all registered handlers for the specified event.
        Args:
            event_name: The name of the event to trigger.
            *args: Positional arguments to pass to event handlers.
            **kwargs: Keyword arguments to pass to event handlers.
        """
        if event_name not in self._event_handlers:
            logger.warning(f"Event handler {event_name} not registered")
            return

        event_handler = self._event_handlers[event_name]

        for handler in event_handler.handlers:
            if self.task_manager:
                task_name = f"{self.name}::{event_name}::{handler.__name__}::handler"
                task = self.task_manager.create_task(
                    self._run_handler(event_name, handler, *args, **kwargs),
                    name=task_name,
                )
                self._event_tasks.add(task)

    def __repr__(self):
        return f"<AgentSession: {self.name} ({len(self.nodes)} nodes)>"



import json
import os
import stat
from pathlib import Path
from typing import Dict, Optional


class AuthClient:
    def __init__(self, app_name: str = "smallestai"):
        self.app_name = app_name
        self.config_dir = Path.home() / f".{app_name}"
        self.credentials_file = self.config_dir / "credentials.json"
        self.config_file = self.config_dir / "config.json"

        self.config_dir.mkdir(parents=True, exist_ok=True)

    def login(self, auth_token: str) -> bool:
        """
        Authenticate user and store credentials
        """
        self._store_credentials(
            {
                "access_token": auth_token,
            }
        )

        return True

    def _store_credentials(self, credentials: Dict) -> None:
        with open(self.credentials_file, "w") as f:
            json.dump(credentials, f, indent=2)

        os.chmod(self.credentials_file, stat.S_IRUSR | stat.S_IWUSR)

    def get_credentials(self) -> Optional[Dict]:
        if not self.credentials_file.exists():
            return None

        try:
            with open(self.credentials_file, "r") as f:
                return json.load(f)
        except Exception:
            return None

    def logout(self) -> None:
        """Remove stored credentials"""
        if self.credentials_file.exists():
            self.credentials_file.unlink()



"""
Audience management.

Usage:
    from smallestai.atoms.helpers import Audience

    audience = Audience()
    audience.create(name="My Audience", phone_numbers=["+1234567890"])
    audience.list()
"""

import os
from typing import Any, Dict, List, Optional, Tuple, Union

import requests

# Default API base URL
DEFAULT_BASE_URL = "https://api.smallest.ai/atoms/v1"


class Audience:
    """
    Manager for audience operations.

    Can be used standalone:
        audience = Audience()
        audience.create(...)
    """

    def __init__(
        self,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
    ):
        """
        Initialize Audience manager.

        Args:
            base_url: API base URL (default: api.smallest.ai)
            api_key: API key (default: SMALLEST_API_KEY env var)
        """
        self.base_url = base_url or os.environ.get("SMALLEST_BASE_URL", DEFAULT_BASE_URL)
        self.api_key = api_key or os.environ.get("SMALLEST_API_KEY", "")

    def _get_headers(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.api_key}"}

    # =========================================================================
    # CRUD Operations
    # =========================================================================

    def create(
        self,
        name: str,
        phone_numbers: List[str],
        names: Optional[List[Tuple[str, str]]] = None,
        description: str = "",
        phone_number_column_name: str = "phoneNumber",
    ) -> Dict[str, Any]:
        """
        Create a new audience.

        Args:
            name: Audience name
            phone_numbers: List of phone numbers
            names: Optional list of (firstName, lastName) tuples
            description: Optional description
            phone_number_column_name: Column name for phone in CSV

        Returns:
            Dict with audience data including _id
        """
        url = f"{self.base_url}/audience"

        # Build CSV content (format: firstName,lastName,phoneNumber)
        csv_lines = ["firstName,lastName,phoneNumber"]
        for i, phone in enumerate(phone_numbers):
            if names and i < len(names):
                first_name, last_name = names[i]
            else:
                first_name, last_name = "User", str(i + 1)
            csv_lines.append(f"{first_name},{last_name},{phone}")

        csv_content = "\n".join(csv_lines)

        files = {"file": ("contacts.csv", csv_content.encode("utf-8"), "text/csv")}
        data = {
            "name": name,
            "phoneNumberColumnName": phone_number_column_name,
            "description": description,
        }

        response = requests.post(url, headers=self._get_headers(), files=files, data=data)
        response.raise_for_status()
        return response.json()  # type: ignore[no-any-return]

    def get(self, audience_id: str) -> Dict[str, Any]:
        """Get audience details."""
        url = f"{self.base_url}/audience/{audience_id}"
        headers = self._get_headers()
        headers["Content-Type"] = "application/json"
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        return response.json()  # type: ignore[no-any-return]

    def get_members(
        self,
        audience_id: str,
        page: int = 1,
        offset: int = 10,
        search: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Get audience members."""
        url = f"{self.base_url}/audience/{audience_id}/members"
        params: Dict[str, Union[str, int]] = {"page": page, "offset": offset}
        if search:
            params["search"] = search

        headers = self._get_headers()
        headers["Content-Type"] = "application/json"
        response = requests.get(url, headers=headers, params=params)
        response.raise_for_status()
        return response.json()  # type: ignore[no-any-return]

    def list(self) -> Dict[str, Any]:
        """List all audiences."""
        url = f"{self.base_url}/audience"
        headers = self._get_headers()
        headers["Content-Type"] = "application/json"
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        return response.json()  # type: ignore[no-any-return]

    def delete(self, audience_id: str) -> Dict[str, Any]:
        """Delete an audience."""
        url = f"{self.base_url}/audience/{audience_id}"
        headers = self._get_headers()
        headers["Content-Type"] = "application/json"
        response = requests.delete(url, headers=headers)
        response.raise_for_status()
        return response.json()  # type: ignore[no-any-return]

    # =========================================================================
    # Contact Management
    # =========================================================================

    def add_contacts(
        self,
        audience_id: str,
        phone_numbers: List[str],
        names: Optional[List[Tuple[str, str]]] = None,
    ) -> Dict[str, Any]:
        """Add contacts to audience."""
        url = f"{self.base_url}/audience/{audience_id}/members"
        headers = self._get_headers()
        headers["Content-Type"] = "application/json"

        # Build members list
        members = []
        for i, phone in enumerate(phone_numbers):
            member: Dict[str, str] = {"phoneNumber": phone}
            if names and i < len(names):
                member["firstName"] = names[i][0]
                member["lastName"] = names[i][1]
            else:
                member["firstName"] = "User"
                member["lastName"] = str(i + 1)
            members.append(member)

        response = requests.post(url, headers=headers, json={"members": members})
        response.raise_for_status()
        return response.json()  # type: ignore[no-any-return]



# This file was auto-generated by Fern from our API Definition.

import contextlib
import typing
from json.decoder import JSONDecodeError

from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.http_response import AsyncHttpResponse, HttpResponse
from ...core.request_options import RequestOptions
from ...core.unchecked_base_model import construct_type
from ..errors.bad_request_error import BadRequestError
from ..errors.internal_server_error import InternalServerError
from ..errors.unauthorized_error import UnauthorizedError
from .types.lightning_request_language import LightningRequestLanguage
from .types.lightning_request_output_format import LightningRequestOutputFormat

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawLightningClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    @contextlib.contextmanager
    def synthesize_lightning_speech(
        self,
        *,
        text: str,
        voice_id: str,
        sample_rate: typing.Optional[int] = OMIT,
        speed: typing.Optional[float] = OMIT,
        language: typing.Optional[LightningRequestLanguage] = OMIT,
        output_format: typing.Optional[LightningRequestOutputFormat] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[HttpResponse[typing.Iterator[bytes]]]:
        """
        Get speech for given text using the Waves API

        Parameters
        ----------
        text : str
            The text to convert to speech.

        voice_id : str
            The voice identifier to use for speech generation.

        sample_rate : typing.Optional[int]
            The sample rate for the generated audio.

        speed : typing.Optional[float]
            The speed of the generated speech.

        language : typing.Optional[LightningRequestLanguage]
            Determines how numbers are spelled out. If set to 'en', numbers will be read as individual digits in English. If set to 'hi', numbers will be read as individual digits in Hindi.

        output_format : typing.Optional[LightningRequestOutputFormat]
            The format of the output audio.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[HttpResponse[typing.Iterator[bytes]]]
            Synthesized speech retrieved successfully.
        """
        with self._client_wrapper.httpx_client.stream(
            "api/v1/lightning/get_speech",
            base_url=self._client_wrapper.get_environment().waves,
            method="POST",
            json={
                "text": text,
                "voice_id": voice_id,
                "sample_rate": sample_rate,
                "speed": speed,
                "language": language,
                "output_format": output_format,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        ) as _response:

            def _stream() -> HttpResponse[typing.Iterator[bytes]]:
                try:
                    if 200 <= _response.status_code < 300:
                        _chunk_size = request_options.get("chunk_size", None) if request_options is not None else None
                        return HttpResponse(
                            response=_response, data=(_chunk for _chunk in _response.iter_bytes(chunk_size=_chunk_size))
                        )
                    _response.read()
                    if _response.status_code == 400:
                        raise BadRequestError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Any,
                                construct_type(
                                    type_=typing.Any,  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    if _response.status_code == 401:
                        raise UnauthorizedError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Any,
                                construct_type(
                                    type_=typing.Any,  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    if _response.status_code == 500:
                        raise InternalServerError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Any,
                                construct_type(
                                    type_=typing.Any,  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(
                        status_code=_response.status_code, headers=dict(_response.headers), body=_response.text
                    )
                raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

            yield _stream()


class AsyncRawLightningClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    @contextlib.asynccontextmanager
    async def synthesize_lightning_speech(
        self,
        *,
        text: str,
        voice_id: str,
        sample_rate: typing.Optional[int] = OMIT,
        speed: typing.Optional[float] = OMIT,
        language: typing.Optional[LightningRequestLanguage] = OMIT,
        output_format: typing.Optional[LightningRequestOutputFormat] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[bytes]]]:
        """
        Get speech for given text using the Waves API

        Parameters
        ----------
        text : str
            The text to convert to speech.

        voice_id : str
            The voice identifier to use for speech generation.

        sample_rate : typing.Optional[int]
            The sample rate for the generated audio.

        speed : typing.Optional[float]
            The speed of the generated speech.

        language : typing.Optional[LightningRequestLanguage]
            Determines how numbers are spelled out. If set to 'en', numbers will be read as individual digits in English. If set to 'hi', numbers will be read as individual digits in Hindi.

        output_format : typing.Optional[LightningRequestOutputFormat]
            The format of the output audio.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[bytes]]]
            Synthesized speech retrieved successfully.
        """
        async with self._client_wrapper.httpx_client.stream(
            "api/v1/lightning/get_speech",
            base_url=self._client_wrapper.get_environment().waves,
            method="POST",
            json={
                "text": text,
                "voice_id": voice_id,
                "sample_rate": sample_rate,
                "speed": speed,
                "language": language,
                "output_format": output_format,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        ) as _response:

            async def _stream() -> AsyncHttpResponse[typing.AsyncIterator[bytes]]:
                try:
                    if 200 <= _response.status_code < 300:
                        _chunk_size = request_options.get("chunk_size", None) if request_options is not None else None
                        return AsyncHttpResponse(
                            response=_response,
                            data=(_chunk async for _chunk in _response.aiter_bytes(chunk_size=_chunk_size)),
                        )
                    await _response.aread()
                    if _response.status_code == 400:
                        raise BadRequestError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Any,
                                construct_type(
                                    type_=typing.Any,  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    if _response.status_code == 401:
                        raise UnauthorizedError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Any,
                                construct_type(
                                    type_=typing.Any,  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    if _response.status_code == 500:
                        raise InternalServerError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Any,
                                construct_type(
                                    type_=typing.Any,  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(
                        status_code=_response.status_code, headers=dict(_response.headers), body=_response.text
                    )
                raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

            yield await _stream()
